/*
 *  Copyright (C) 2003 Patrizio Bruno.
 *  E-mail: desertconsulting@firenze.net.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include <math.h>
#include <gtk/gtk.h>

#include "gtkmdi.h"
#include "gtkmdi_internal.h"
#include "mdi-marshal.h"

guint gtk_mdi_signals[LAST_SIGNAL] = {0};

static void gtk_mdi_class_init (GtkMdiClass *);
static void gtk_mdi_init (GtkMdi *);
static void gtk_mdi_finalize (GObject *);
static void gtk_mdi_destroy (GtkObject *);
static void gtk_mdi_size_request (GtkWidget *widget,
				GtkRequisition *requisition);
static void gtk_mdi_size_allocate (GtkWidget *widget,
				GtkAllocation *allocation);
static gboolean gtk_mdi_expose (GtkWidget *, GdkEventExpose *);
static void gtk_mdi_realize (GtkWidget *widget);
static void gtk_mdi_unrealize (GtkWidget *widget);
static gboolean gtk_mdi_real_focus (GtkWidget *widget,
				GtkDirectionType direction);
static void gtk_mdi_state_changed (GtkWidget *widget,
				GtkStateType previous_state);
static void gtk_mdi_style_set (GtkWidget *widget, GtkStyle *previous_style);
static void gtk_mdi_add (GtkContainer *container, GtkWidget *widget);
static void gtk_mdi_remove (GtkContainer *container, GtkWidget *widget);
static void gtk_mdi_set_scroll_adjustments (GtkLayout *layout,
					GtkAdjustment *arg1,
					GtkAdjustment *arg2);
static gboolean gtk_mdi_button_press (GtkWidget *, GdkEventButton *);
static gboolean gtk_mdi_button_release (GtkWidget *, GdkEventButton *);
static gboolean gtk_mdi_button_motion (GtkWidget *, GdkEventMotion *);
static GtkLayoutChild *gtk_mdi_find_window(GtkMdi *mdi, gint x, gint y);
static void gtk_mdi_clamp_coords_for_window(GtkMdi *mdi, gint *x, gint *y);

static gint _gtk_mdi_compare_window (gconstpointer a, gconstpointer b);

static GtkLayoutClass *parent_class;

struct _GtkMdiPrivate
{
	GtkMdiWindowId current_id;
	GSList *winlist;
	gint x0;
	gint y0;
	GdkCursor *cross_cursor;
	GdkCursor *arrow_cursor;
	gboolean resizing;
};

GType
gtk_mdi_get_type (void)
{
	static GType gtk_mdi_type = 0;

	if (!gtk_mdi_type)
	{
		static const GTypeInfo gtk_mdi_info = {
			sizeof (GtkMdiClass),
			NULL,
			NULL,
			(GClassInitFunc) gtk_mdi_class_init,
			NULL,
			NULL,
			sizeof (GtkMdi),
			0,
			(GInstanceInitFunc) gtk_mdi_init
		};

		gtk_mdi_type = g_type_register_static (GTK_TYPE_LAYOUT,
							"GtkMdi",
							&gtk_mdi_info, 0);
	}

	return gtk_mdi_type;
}

static void
gtk_mdi_class_init (GtkMdiClass *class)
{
	GObjectClass *o_class;
	GtkObjectClass *gtko_class;
	GtkWidgetClass *w_class;
	GtkContainerClass *c_class;
	GtkLayoutClass *l_class;

	parent_class = g_type_class_peek_parent (class);
	o_class = (GObjectClass *) class;
	gtko_class = (GtkObjectClass *) class;
	w_class = (GtkWidgetClass *) class;
	c_class = (GtkContainerClass *) class;

	gtk_mdi_signals[WINDOW_OPENED] = g_signal_new ("window_opened",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_opened),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	gtk_mdi_signals[WINDOW_CLOSED] = g_signal_new ("window_closed",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_closed),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__INT,
					G_TYPE_NONE, 1,
					G_TYPE_INT);

	gtk_mdi_signals[WINDOW_MAXIMIZED] = g_signal_new ("window_maximized",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_maximized),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	gtk_mdi_signals[WINDOW_UNMAXIMIZED] = g_signal_new ("window_unmaximized",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_unmaximized),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	gtk_mdi_signals[WINDOW_ICONIFIED] = g_signal_new ("window_iconified",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_iconified),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	gtk_mdi_signals[WINDOW_DEICONIFIED] = g_signal_new ("window_deiconified",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_deiconified),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	gtk_mdi_signals[WINDOW_PROPERTY_CHANGED] = g_signal_new (
					"window_property_changed",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
						window_property_changed),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT_STRING_BOOLEAN_BOOLEAN,
					G_TYPE_NONE, 4,
					GTK_MDI_WINDOW_TYPE,
					G_TYPE_STRING,
					G_TYPE_BOOLEAN,
					G_TYPE_BOOLEAN);

	gtk_mdi_signals[ICON_CHANGED] = g_signal_new ("window_icon_changed",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_icon_changed),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	gtk_mdi_signals[TITLE_CHANGED] = g_signal_new ("window_title_changed",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_title_changed),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	gtk_mdi_signals[FOCUS_GET] = g_signal_new ("window_focus_get",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_focus_get),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	gtk_mdi_signals[WINDOW_RESIZED] = g_signal_new ("window_resized",
					G_TYPE_FROM_CLASS (o_class),
					G_SIGNAL_RUN_FIRST,
					G_STRUCT_OFFSET (GtkMdiClass,
							window_resized),
					NULL, NULL,
					g_cclosure_user_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					GTK_MDI_WINDOW_TYPE);

	o_class->finalize = gtk_mdi_finalize;
	//gtko_class->destroy = gtk_mdi_destroy;

	//w_class->size_request = gtk_mdi_size_request;
	w_class->size_allocate = gtk_mdi_size_allocate;
	w_class->expose_event = gtk_mdi_expose;
	w_class->realize = gtk_mdi_realize;
	w_class->unrealize = gtk_mdi_unrealize;
	w_class->button_press_event = gtk_mdi_button_press;
	//w_class->motion_notify_event = gtk_mdi_button_motion;
	w_class->button_release_event = gtk_mdi_button_release;
	//w_class->focus = gtk_mdi_real_focus;
	//w_class->state_changed = gtk_mdi_state_changed;
	//w_class->style_set = gtk_mdi_style_set;

	//c_class->add = gtk_mdi_add;
	//c_class->remove = gtk_mdi_remove;

	//l_class->set_scroll_adjustments = gtk_mdi_set_scroll_adjustments;
}

static void
gtk_mdi_destroy (GtkObject *object)
{
	g_return_if_fail (GTK_IS_MDI (object));

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		GTK_OBJECT_CLASS (parent_class)->destroy (object);

}

static void
gtk_mdi_size_request (GtkWidget *widget, GtkRequisition *requisition)
{
	gint x;
	gint y;

	g_return_if_fail (GTK_IS_MDI (widget));
	g_return_if_fail (requisition != NULL);

	gdk_drawable_get_size (gtk_widget_get_parent_window (widget), &x, &y);
	requisition->width = x;
	requisition->height = y;
}

static void
gtk_mdi_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
	g_return_if_fail (GTK_IS_MDI (widget));
	g_return_if_fail (allocation != NULL);

	GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);
}

static gboolean
gtk_mdi_expose (GtkWidget *widget, GdkEventExpose *event)
{
	g_return_val_if_fail (GTK_IS_MDI (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	if (GTK_WIDGET_CLASS (parent_class)->expose_event)
		(GTK_WIDGET_CLASS (parent_class)->expose_event) (widget,
								event);

	return TRUE;
}

static void
gtk_mdi_realize (GtkWidget *widget)
{
	g_return_if_fail (GTK_IS_MDI (widget));

	if (GTK_WIDGET_CLASS (parent_class)->realize)
		GTK_WIDGET_CLASS (parent_class)->realize (widget);

	widget->style = gtk_style_attach (widget->style, widget->window);

	gdk_window_set_background (GTK_LAYOUT (widget)->bin_window,
			&(GTK_WIDGET (widget)->style->base [GTK_STATE_NORMAL]));
}

static void
gtk_mdi_unrealize (GtkWidget *widget)
{
	g_return_if_fail (GTK_IS_MDI (widget));

	GTK_WIDGET_CLASS (parent_class)->unrealize (widget);
}

static gboolean
gtk_mdi_real_focus (GtkWidget *widget, GtkDirectionType direction)
{
	if (GTK_WIDGET_CAN_FOCUS (widget))
	{
		GTK_WIDGET_UNSET_FLAGS (widget, GTK_CAN_FOCUS);
	}
	return GTK_WIDGET_CLASS (parent_class)->focus (widget, direction);
}

static void
gtk_mdi_state_changed (GtkWidget *widget, GtkStateType previous_state)
{
}

static void
gtk_mdi_style_set (GtkWidget *widget, GtkStyle *previous_style)
{
}

static void
gtk_mdi_add (GtkContainer *container, GtkWidget *widget)
{
	g_return_if_fail (GTK_IS_MDI (container));
	g_return_if_fail (GTK_IS_MDI_WINDOW (widget));

	gtk_mdi_add_window (GTK_MDI (container), GTK_MDI_WINDOW (widget));
}

static void
gtk_mdi_remove (GtkContainer *container, GtkWidget *widget)
{
	g_return_if_fail (GTK_IS_MDI (container));
	g_return_if_fail (GTK_IS_MDI_WINDOW (widget));

	gtk_mdi_close_window (GTK_MDI (container),
			GTK_MDI_WINDOW (widget)->id);

	if (GTK_CONTAINER_CLASS (parent_class)->remove)
		(GTK_CONTAINER_CLASS (parent_class)->remove) (container,
								widget);
}

static void
gtk_mdi_set_scroll_adjustments (GtkLayout *layout, GtkAdjustment *arg1,
				GtkAdjustment *arg2)
{
	g_return_if_fail (GTK_IS_MDI (layout));
	g_return_if_fail (GTK_IS_ADJUSTMENT (arg1));
	g_return_if_fail (GTK_IS_ADJUSTMENT (arg2));

	if (GTK_LAYOUT_CLASS (parent_class)->set_scroll_adjustments)
		GTK_LAYOUT_CLASS (parent_class)->set_scroll_adjustments (layout, arg1, arg2);
}

static void
gtk_mdi_init (GtkMdi *mdi)
{
	mdi->private = g_new0 (GtkMdiPrivate, 1);
	mdi->private->winlist = g_slist_alloc ();
	mdi->private->current_id = 0;
	mdi->private->cross_cursor = gdk_cursor_new(GDK_PLUS);
	mdi->private->arrow_cursor = gdk_cursor_new(GDK_ARROW);
}

static void
gtk_mdi_finalize (GObject *object)
{
	if (G_OBJECT_CLASS (parent_class)->finalize)
		(G_OBJECT_CLASS (parent_class)->finalize) (object);
}

GtkWidget *
gtk_mdi_new (void)
{
	GtkMdi *new_mdi;

	new_mdi = (GtkMdi*)gtk_widget_new (GTK_MDI_TYPE,
					"hadjustment", NULL,
                                        "vadjustment", NULL, NULL);
	gtk_widget_add_events(GTK_WIDGET (new_mdi), GDK_BUTTON_PRESS_MASK
						| GDK_BUTTON_RELEASE_MASK
						| GDK_BUTTON_MOTION_MASK
						| GDK_POINTER_MOTION_MASK);
	gtk_layout_set_size (GTK_LAYOUT (new_mdi), 1600, 1200);

	return GTK_WIDGET (new_mdi);
}

void
gtk_mdi_add_window (GtkMdi *mdi, GtkMdiWindow *window)
{
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	gtk_layout_put (GTK_LAYOUT (mdi), GTK_WIDGET (window), 1, 1);
	window->id = (mdi->private->current_id)++;
	window->mdicontainer = mdi;
	g_slist_append(mdi->private->winlist, window);
	g_signal_emit (mdi, gtk_mdi_signals[WINDOW_OPENED], 0, window);
}

void
gtk_mdi_close_window (GtkMdi *mdi, GtkMdiWindowId id)
{
	GtkMdiWindow *window;

	window = gtk_mdi_get_window (mdi, id);
	gtk_mdi_window_close (window);
}

GtkMdiWindow *
gtk_mdi_get_window (GtkMdi *mdi, GtkMdiWindowId id)
{
	GtkMdiWindow *window = NULL;
	GSList *element;

	window = GTK_MDI_WINDOW (gtk_mdi_window_new ());

	element = g_slist_find_custom (mdi->private->winlist, window,
					_gtk_mdi_compare_window);
	g_object_unref (window);
	g_free (window);
	window = NULL;

	if (element && element->data) window = element->data;

	return window;
}

static gint
_gtk_mdi_compare_window (gconstpointer a, gconstpointer b)
{
	gint ret = 0;

	if(a && b)
	{
		GtkMdiWindow *win_a = (GtkMdiWindow *)a;
		GtkMdiWindow *win_b = (GtkMdiWindow *)b;

		ret = win_a->id - win_b->id;
	}

	return ret;
}

void
gtk_mdi_maximize_window (GtkMdi *mdi, GtkMdiWindowId id)
{
	GtkMdiWindow *window;

	window = gtk_mdi_get_window (mdi, id);
	gtk_mdi_window_maximize (window);
}

void
gtk_mdi_unmaximize_window (GtkMdi *mdi, GtkMdiWindowId id)
{
	GtkMdiWindow *window;

	window = gtk_mdi_get_window (mdi, id);
	gtk_mdi_window_unmaximize (window);
}

void
gtk_mdi_iconify_window (GtkMdi *mdi, GtkMdiWindowId id)
{
	GtkMdiWindow *window;

	window = gtk_mdi_get_window (mdi, id);
	gtk_mdi_window_iconify (window);
}

void
gtk_mdi_deiconify_window (GtkMdi *mdi, GtkMdiWindowId id)
{
	GtkMdiWindow *window;

	window = gtk_mdi_get_window (mdi, id);
	gtk_mdi_window_deiconify (window);
}

static gboolean
gtk_mdi_button_motion (GtkWidget *widget, GdkEventMotion *event)
{
	GtkMdi *mdi = GTK_MDI (widget);
	gint x = 0;
	gint y = 0;
	g_return_val_if_fail (GTK_IS_MDI (widget), FALSE);

	if (mdi->active)
	{
		gdk_window_get_position(event->window, &x, &y);
		x += (gint)event->x;
		y += (gint)event->y;
		gtk_mdi_clamp_coords_for_window(mdi, &x, &y);
		if (x != mdi->private->x0 || y != mdi->private->y0)
		{
			x -= mdi->private->x0 -
				GTK_WIDGET (mdi->active)->allocation.x;
			y -= mdi->private->y0 -
				GTK_WIDGET (mdi->active)->allocation.y;
			if((x > 0) && (y > 0))
			gtk_layout_move(GTK_LAYOUT(mdi),
					GTK_WIDGET (mdi->active), x, y);

			//gdk_window_get_position(event->window, &x, &y);
			//x += (gint)event->x;
			//y += (gint)event->y;
			GTK_MDI (widget)->private->x0 = x;
			GTK_MDI (widget)->private->y0 = y;
		}
	}

	return FALSE;
}

static gboolean
gtk_mdi_button_press (GtkWidget *widget, GdkEventButton *event)
{
	GtkLayoutChild *child;
	GtkAllocation *allocation;
	gint x;
	gint y;

	gdk_window_get_position(event->window, &x, &y);
	printf ("x:%d - y:%d\n", x, y);
	x += (gint)event->x;
	y += (gint)event->y;

	child = gtk_mdi_find_window (GTK_MDI (widget), x, y);
	if (child && GTK_IS_MDI_WINDOW (child->widget))
			allocation = &(child->widget->allocation);

	if (child && GTK_IS_MDI_WINDOW (child->widget) &&
		x >= allocation->x + allocation->width - 8
		&& x < allocation->x + allocation->width
		&& y >= allocation->y + allocation->height - 8
		&& y < allocation->y + allocation->height)

	{
		GTK_MDI (widget)->private->resizing = TRUE;
	}

	if (child && GTK_IS_MDI_WINDOW (child->widget))
	{
		gdk_window_set_cursor(GTK_LAYOUT(widget)->bin_window,
				GTK_MDI (widget)->private->cross_cursor);

		GTK_MDI (widget)->active = GTK_MDI_WINDOW (child->widget);
		GTK_MDI (widget)->private->x0 = x;
		GTK_MDI (widget)->private->y0 = y;
		gtk_container_set_focus_child (GTK_CONTAINER (widget),
					child->widget);
		gtk_layout_freeze (GTK_LAYOUT (widget));
		gtk_mdi_window_grab_focus (child->widget);
	}

	return TRUE;
}

static gboolean
gtk_mdi_button_release (GtkWidget *widget, GdkEventButton *event)
{
	GtkMdi *mdi = GTK_MDI (widget);
	GtkLayoutChild *child;
	gint x;
	gint y;
	gint w;
	gint h;

	g_return_val_if_fail (GTK_IS_MDI (mdi), FALSE);
	g_return_val_if_fail (GTK_IS_MDI_WINDOW (mdi->active), FALSE);

	gdk_window_get_position (event->window, &x, &y);
	x += (gint)event->x;
	y += (gint)event->y;
	w = x - mdi->private->x0;
	h = y - mdi->private->y0;
	gtk_mdi_clamp_coords_for_window (mdi, &x, &y);
	if (x != mdi->private->x0 || y != mdi->private->y0)
	{
		x -= mdi->private->x0 - GTK_WIDGET (mdi->active)->allocation.x;
		y -= mdi->private->y0 - GTK_WIDGET (mdi->active)->allocation.y;
		if (mdi->private->resizing)
			gtk_mdi_window_resize (mdi->active,
				GTK_WIDGET (mdi->active)->allocation.width + w,
				GTK_WIDGET (mdi->active)->allocation.height + h);
		else gtk_layout_move(GTK_LAYOUT(mdi),
				GTK_WIDGET (mdi->active), x, y);

		mdi->private->resizing = FALSE;
	}

	gdk_window_set_cursor(GTK_LAYOUT(widget)->bin_window,
				GTK_MDI (widget)->private->arrow_cursor);
	mdi->active = NULL;
	gtk_layout_thaw (GTK_LAYOUT (mdi));

	return TRUE;
}

static GtkLayoutChild *
gtk_mdi_find_window(GtkMdi *mdi, gint x, gint y)
{
	GList *chpl;
	GtkLayoutChild *ret = NULL;

	for (chpl = GTK_LAYOUT(mdi)->children; chpl;
		chpl = g_list_next(chpl))
	{
		GtkLayoutChild *child;
		GtkAllocation *allocation;
	
		child = ((GtkLayoutChild *)(chpl->data));
		if (GTK_IS_MDI_WINDOW (child->widget) &&
			((mdi->active && (child->widget == mdi->active)) ||
		(gtk_mdi_window_get_state (GTK_MDI_WINDOW (child->widget))
				!= GTK_MDI_WINDOW_ICONIFIED)))
		{
			allocation = &(child->widget->allocation);
			if (x >= allocation->x
				&& x < (allocation->x + allocation->width)
				&& y >= allocation->y
				&& y < (allocation->y + 20))//allocation->height))
			{
				ret = child;
				gtk_mdi_window_set_has_focus (
						GTK_MDI_WINDOW (child->widget),
						FALSE);
			}
		}
	}

	if (ret) gtk_mdi_window_grab_focus (GTK_MDI_WINDOW (ret->widget));

	return ret;
}

static void
gtk_mdi_clamp_coords_for_window(GtkMdi *mdi, gint *x, gint *y)
{
	GtkAllocation *allocation;
	gint min, max;

	allocation = &(GTK_WIDGET (mdi->active)->allocation);

	min = mdi->private->x0 - allocation->x;
	max = GTK_WIDGET(mdi)->allocation.width -
				(allocation->width - min) - 1;
	*x = CLAMP(*x, min, max);

	min = mdi->private->y0 - allocation->y;
	max = GTK_WIDGET(mdi)->allocation.height -
				(allocation->height - min) - 1;
	*y = CLAMP(*y, min, max);
}

